#include "stdafx.h"
#include "lb041227_mfc.h"
#include "lb041227_mfcDlg.h"
#include <windowsx.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

//   
const DWORD ID_NEWGAME	= 102;
const DWORD ID_EXIT		= 103;
const DWORD ID_RULES	= 104;

//    
typedef enum { GBCS_NORMAL = 0, GBCS_SELECTED, GBCS_DISABLED } _GBCS;

//   
typedef struct {
	int number;					//    (0 -   )
	_GBCS state;				//  
} _GameBoard;

const int BOARD_WIDTH	= 10;	//      -
const int BOARD_HEIGHT	= 10;	//      -
const int CELL_SIZE		= 30;	//      ()
const int BACK_SIZE		= 3;	// - 
int CursorX, CursorY;			//     
int SelectX, SelectY;			//    
int Summ;						//    
HBRUSH hbrNormal;				//     
HBRUSH hbrSelected;				//     
HBRUSH hbrBackground;			//     
HBRUSH hbrCursor;				//  
HFONT hfntCell;					//  
SYSTEMTIME	stStart;			//   
bool GameComplete;				//   
int VictoryCount;				//  
HDC hMemDC;						//     
HBITMAP hMemBitmap;				//     
HBITMAP hMemBitmapOld;			//       
HBITMAP hbmpImg [BACK_SIZE];	// 
const LPCTSTR szTitle = " \"\"";		//  
_GameBoard GameBoard [BOARD_HEIGHT][BOARD_WIDTH];	//  

int GetGameTime ();
bool TestCellH (int x, int y);
bool TestCellV (int x, int y);
bool FillGameBoardCell (int x, int y);
void FillGameBoard ();
void DrawCell (int x, int y);
void DrawGameBoard ();
void DrawBackground (HDC hDC, const RECT* rc);
bool ClickCell (int x, int y);

//   
int GetGameTime ()
{
	//   
	SYSTEMTIME stCurrent;
	GetLocalTime (&stCurrent);
	//        
    int elapsed = 
		(stCurrent.wSecond + (stCurrent.wMinute + stCurrent.wHour * 60) * 60) -
		(stStart.wSecond + (stStart.wMinute + stStart.wHour * 60) * 60);
	if (elapsed < 0)
		//  -   
		elapsed = 24 * 60 * 60 + elapsed;
	return elapsed;
}

//     
void DrawCell (int x, int y)
{
	//       
	RECT rc = { x * CELL_SIZE, y * CELL_SIZE, (x + 1) * CELL_SIZE, (y + 1) * CELL_SIZE };
	DrawBackground (hMemDC, &rc);
	switch (GameBoard [y][x].state) {
		case GBCS_NORMAL:
		{
			HBRUSH hbrOld = SelectBrush (hMemDC, hbrNormal);
			Rectangle (hMemDC, rc.left, rc.top, rc.right - 1, rc.bottom - 1);
			SelectBrush (hMemDC, hbrOld);
			break;
		}
		case GBCS_SELECTED:
		{
			HBRUSH hbrOld = SelectBrush (hMemDC, hbrSelected);
			Rectangle (hMemDC, rc.left, rc.top, rc.right - 1, rc.bottom - 1);
			SelectBrush (hMemDC, hbrOld);
			break;
		}
	}
	//  
	if (GameBoard [y][x].state != GBCS_DISABLED) {
		char txt [33];
		wsprintf (txt, "%d", GameBoard [y][x].number);
		SetTextColor (hMemDC, RGB (0, 0, 0));
		int old = SetBkMode (hMemDC, TRANSPARENT);
		HFONT hfntOld = SelectFont (hMemDC, hfntCell);
		DrawText (hMemDC, txt, -1, &rc, DT_CENTER | DT_VCENTER | DT_SINGLELINE);
		SelectFont (hMemDC, hfntOld);
		SetBkMode (hMemDC, old);
	}
}

//  
void DrawBackground (HDC hDC, const RECT* rc)
{
	HDC hTmpDC = CreateCompatibleDC (hDC);
	HBITMAP hTmpBitmapOld = SelectBitmap (hTmpDC, hbmpImg [VictoryCount % BACK_SIZE]);
	FillRect (hDC, rc, hbrBackground);
	BitBlt (hDC, rc->left, rc->top, rc->right - rc->left, rc->bottom - rc->top,
		hTmpDC, rc->left, rc->top, SRCCOPY); 
	SelectBitmap (hTmpDC, hTmpBitmapOld);
	DeleteDC (hTmpDC);
}

//       
void DrawGameBoard ()
{
	//   
	const RECT rc = { 0, 0, BOARD_HEIGHT * CELL_SIZE - 1, BOARD_WIDTH * CELL_SIZE - 1 };
	DrawBackground (hMemDC, &rc);
	//   
	for (int y = 0; y < BOARD_HEIGHT; ++y)
		for (int x = 0; x < BOARD_WIDTH; ++x)
			DrawCell (x, y);
}

//      
bool ClickCell (int x, int y)
{
	if (SelectX == -1) {
		if (GameBoard [y][x].state == GBCS_NORMAL) {
			//   
			GameBoard [y][x].state = GBCS_SELECTED;
			SelectX = x;
			SelectY = y;
			DrawCell (x, y);
		}
	} else {
		if (SelectX == x && SelectY == y) {
			//    
			GameBoard [y][x].state = GBCS_NORMAL;
			SelectX = -1;
			SelectY = -1;
			DrawCell (x, y);
		} else 
			if (GameBoard [y][x].state == GBCS_NORMAL &&
				((SelectX == x && (SelectY == y + 1 || SelectY == y - 1)) ||
				((SelectX == x + 1 || SelectX == x - 1) && SelectY == y)) &&
				(GameBoard [y][x].number + GameBoard [SelectY][SelectX].number == 10)) {
				//   
				GameBoard [y][x].state = GBCS_DISABLED;
				GameBoard [SelectY][SelectX].state = GBCS_DISABLED;
				DrawCell (x, y);
				DrawCell (SelectX, SelectY);
				SelectX = -1;
				SelectY = -1;
				Summ += 2;
			}
	}
	return (Summ == BOARD_WIDTH * BOARD_HEIGHT);
}

//    
bool TestCellH (int x, int y)
{
	GameBoard [y][x].number = (rand () * 5) / RAND_MAX + 1;
	GameBoard [y][x + 1].number = 10 - GameBoard [y][x].number;
	//    
	if (x + 1 == BOARD_WIDTH) {
		if (y + 1 == BOARD_HEIGHT)
			return true;
		else
			if (FillGameBoardCell (0, y + 1))
				return true;
	} else
		if (FillGameBoardCell (x + 1, y))
			return true;
	//  
	GameBoard [y][x].number = 0;
	GameBoard [y][x + 1].number = 0;
	return false;
}

//    
bool TestCellV (int x, int y)
{
	GameBoard [y][x].number = (rand () * 5) / RAND_MAX + 1;		// 1-5
	GameBoard [y + 1][x].number = 10 - GameBoard [y][x].number;
	//    
	if (x + 1 == BOARD_WIDTH) {
		if (y + 1 == BOARD_HEIGHT)
			return true;
		else
			if (FillGameBoardCell (0, y + 1))
				return true;
	} else
		if (FillGameBoardCell (x + 1, y))
			return true;
	//  
	GameBoard [y][x].number = 0;
	GameBoard [y + 1][x].number = 0;
	return false;
}

//      
bool FillGameBoardCell (int x, int y)
{
	//      (x,y) 
	if (GameBoard [y][x].number == 0) {
		//       
		bool h = (x + 1 != BOARD_WIDTH  && GameBoard [y][x + 1].number == 0);
		bool v = (y + 1 != BOARD_HEIGHT && GameBoard [y + 1][x].number == 0);
		//    , ,
		//     ,   
		if ((h && !v) || (h && v && ((rand () * 2) / RAND_MAX) == 1)) {
			//    
			if (TestCellH (x, y))
				return true;
			h = false;
		}
		//   
		if (v) {
			//    
			if (TestCellV (x, y))
				return true;
			if (h)
				//    
				if (TestCellH (x, y))
					return true;
		}
		//   
		return false;
	}
	//    
	if (x + 1 == BOARD_WIDTH) {
		if (y + 1 == BOARD_HEIGHT)
			return true;
		else
			return FillGameBoardCell (0, y + 1);
	} else
		return FillGameBoardCell (x + 1, y);
}

//       
void FillGameBoard ()
{
	CursorX = -1;
	CursorY = -1;
	SelectX = -1;
	SelectY = -1;
	Summ = 0;
	//     
	srand (GetTickCount ());
	ZeroMemory (GameBoard, sizeof (GameBoard));
	FillGameBoardCell (0, 0);
	//  
	DrawGameBoard ();
	//    
	GetLocalTime (&stStart);
	GameComplete = false;
}

BEGIN_MESSAGE_MAP(Clb041227_mfcDlg, CDialog)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_WM_TIMER()
	ON_WM_LBUTTONDOWN()
	ON_WM_DESTROY()
END_MESSAGE_MAP()

BOOL Clb041227_mfcDlg::OnInitDialog()
{
	CDialog::OnInitDialog();
	m_hIcon = LoadIcon (NULL, IDI_ASTERISK);
	SetIcon (m_hIcon, TRUE);
	SetIcon (m_hIcon, FALSE);

	//    
	const DWORD dwStyle = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU |
		WS_MINIMIZEBOX | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_VISIBLE;
	const DWORD dwExStyle = WS_EX_APPWINDOW;
	RECT rc = {0, 0, BOARD_HEIGHT * CELL_SIZE - 1, BOARD_WIDTH * CELL_SIZE - 1};
	AdjustWindowRectEx (&rc, dwStyle, TRUE, dwExStyle);
	SetWindowPos (NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top,
		SWP_NOMOVE | SWP_NOZORDER);

	HMENU hMenu = CreateMenu ();		
	AppendMenu (hMenu, 0, ID_RULES, "& ");
	AppendMenu (hMenu, 0, ID_NEWGAME, "&  ");
	AppendMenu (hMenu, 0, ID_EXIT, "&  ");
	SetMenu (CMenu::FromHandle (hMenu));

	for (int i = 0; i < BACK_SIZE; ++i)
		hbmpImg [i] = (HBITMAP) LoadImage (AfxGetInstanceHandle (), MAKEINTRESOURCE (IDB_IMG1 + i),
			IMAGE_BITMAP, 300, 300, LR_CREATEDIBSECTION);
	//    
	hbrNormal = CreateSolidBrush (RGB (255, 255, 255));
	hbrSelected = CreateHatchBrush (HS_DIAGCROSS, RGB (128, 128, 255));
	hbrBackground = CreateSolidBrush (RGB (128, 128, 255));
	hbrCursor = CreateSolidBrush (RGB (0, 0, 255));
	//    
	LOGFONT lf;
	ZeroMemory (&lf, sizeof (lf));
	lf.lfHeight = - CELL_SIZE;
	lf.lfQuality = ANTIALIASED_QUALITY;
	lstrcpy (lf.lfFaceName, "Arial");
	hfntCell = CreateFontIndirect (&lf);
	//     
	HDC hDC = ::GetDC (GetSafeHwnd ());
	hMemDC = CreateCompatibleDC (hDC);
	hMemBitmap = CreateCompatibleBitmap (hDC,
		BOARD_HEIGHT * CELL_SIZE, BOARD_WIDTH * CELL_SIZE);
	hMemBitmapOld = SelectBitmap (hMemDC, hMemBitmap);
	::ReleaseDC (GetSafeHwnd (), hDC);
	VictoryCount = 0;
	//    
	FillGameBoard ();
	//   
	SetTimer (100, 1000, NULL);
	
	return TRUE;
}

void Clb041227_mfcDlg::OnPaint() 
{
	if (IsIconic()) {
		CPaintDC dc(this);
		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;
		dc.DrawIcon(x, y, m_hIcon);
	} else {
		CPaintDC dc(this);
		//   
		::BitBlt (dc.GetSafeHdc (), 0, 0, BOARD_WIDTH * CELL_SIZE, BOARD_HEIGHT * CELL_SIZE,
			hMemDC, 0, 0, SRCCOPY);
	}
}

HCURSOR Clb041227_mfcDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void Clb041227_mfcDlg::OnTimer(UINT nIDEvent)
{
	if (GameComplete) {
		SetWindowText (szTitle);
	} else {
		int elapsed = GetGameTime ();
		char txt [256];
		wsprintf (txt, "%s - %02d:%02d", szTitle, elapsed / 60, elapsed % 60);
		SetWindowText (txt);
	}
	CDialog::OnTimer(nIDEvent);
}

void Clb041227_mfcDlg::OnLButtonDown(UINT nFlags, CPoint point)
{
	//     
	int x = point.x / CELL_SIZE; 
	int y = point.y / CELL_SIZE; 
	GameComplete = ClickCell (x, y);
	InvalidateRect (NULL, FALSE);
	UpdateWindow ();
	if (GameComplete) {
		//  
		int elapsed = GetGameTime ();
		char txt [256];
		wsprintf (txt, "!   %02d:%02d. "
			" ! :-) ...  .\n"
			"      ...",
			elapsed / 60, elapsed % 60);
		MessageBox (txt, szTitle, MB_OK | MB_ICONEXCLAMATION);
		++VictoryCount;
		FillGameBoard ();
		InvalidateRect (NULL, TRUE);
		UpdateWindow ();
	}
	CDialog::OnLButtonDown(nFlags, point);
}

BOOL Clb041227_mfcDlg::OnCommand(WPARAM wParam, LPARAM lParam)
{
	switch (LOWORD (wParam)) {
		case ID_RULES:
			//    " "
			MessageBox (" :\n"
				"     ,     .\n"
				"      , "
				"    10.\n"
				"!    ,     . :-)\n"
				"    ,     ...",
				szTitle, MB_OK | MB_ICONINFORMATION);
			return 1;
		case ID_NEWGAME:
			//    "  "
			FillGameBoard ();
			InvalidateRect (NULL, TRUE);
			UpdateWindow ();
			return 1;
		case ID_EXIT:
			//    "  "
			DestroyWindow ();
			return 1;
	}
	return CDialog::OnCommand(wParam, lParam);
}

void Clb041227_mfcDlg::OnDestroy()
{
	//  ,  
	KillTimer (100);
	DeleteBrush (hbrNormal);
	DeleteBrush (hbrSelected);
	DeleteBrush (hbrBackground);
	DeleteBrush (hbrCursor);
	DeleteFont (hfntCell);
	SelectBitmap (hMemDC, hMemBitmapOld);
	DeleteDC (hMemDC);
	DeleteBitmap (hMemBitmap);
	for (int i = 0; i < BACK_SIZE; ++i)
		DeleteBitmap (hbmpImg [i]);
	CDialog::OnDestroy();
}
